// Windows and Windows DLLs
#include <windows.h>
#include <Xinput.h>
#include <dsound.h>

// Local imports
#include "win32_handmade.h"
#include "handmade.cpp"

global ATOM HH_CTRLW;
global bool globalRunning;
global Win32OffscreenBuffer globalBackBuffer;
global LPDIRECTSOUNDBUFFER globalSecondaryBuffer;
global int64 globalPerfCountFrequency;

// XInputGetState
#define X_INPUT_GET_STATE(name) DWORD WINAPI name(DWORD dwUserIndex, XINPUT_STATE *pState)
typedef X_INPUT_GET_STATE(XInputGetStateFn);
X_INPUT_GET_STATE(XInputGetStateStub) { return ERROR_DEVICE_NOT_CONNECTED; }
global XInputGetStateFn *XInputGetStateDyn = XInputGetStateStub;
#define XInputGetState XInputGetStateDyn

// XInputSetState
#define X_INPUT_SET_STATE(name) DWORD WINAPI name(DWORD dwUserIndex, XINPUT_VIBRATION *pVibration)
typedef X_INPUT_SET_STATE(XInputSetStateFn);
X_INPUT_SET_STATE(XInputSetStateStub) { return ERROR_DEVICE_NOT_CONNECTED; }
global XInputSetStateFn *XInputSetStateDyn = XInputSetStateStub;
#define XInputSetState XInputSetStateDyn

typedef HRESULT WINAPI DirectSoundCreateFn(LPCGUID pcGuidDevice, LPDIRECTSOUND *ppDS, LPUNKNOWN pUnkOuter);

#define stackAlloc(size, type) (type*)_alloca(size*sizeof(type))

void DEBUG_platformFreeFileMemory(void *fileMemory) {
    if (fileMemory) {
        VirtualFree(fileMemory, NULL, MEM_RELEASE);
    }
}

DebugReadFileResult DEBUG_platformReadEntireFile(char *filename) {
    DebugReadFileResult result = {};
    HANDLE fileHandle = CreateFileA(filename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, NULL, NULL);
    if (fileHandle != INVALID_HANDLE_VALUE) {
        LARGE_INTEGER fileSize;
        if (GetFileSizeEx(fileHandle, &fileSize)) {
            uint32 fileSize32 = safeTruncateUInt64(fileSize.QuadPart);
            result.contents = VirtualAlloc(NULL, fileSize.QuadPart, MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE);
            if (result.contents) {
                DWORD bytesRead;
                if (ReadFile(fileHandle, result.contents, (DWORD)fileSize.QuadPart, &bytesRead, NULL) && (fileSize32 == (uint32)bytesRead)) {
                    result.contentsSize = fileSize32;
                } else {
                    DEBUG_platformFreeFileMemory(result.contents);
                    result.contents = NULL;
                    // logging
                }
            } else {
                result.contents = NULL;
                // logging
            }
        } else {
            // logging
        }
        CloseHandle(fileHandle);
    } else {
        // logging
    }
    return result;
}

bool32 DEBUG_platformWriteEntireFile(char *filename, uint32 memorySize, void *memory) {
    bool32 result = false;
    HANDLE fileHandle = CreateFileA(filename, GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, NULL, NULL);
    if (fileHandle != INVALID_HANDLE_VALUE) {
        DWORD bytesWritten;
        if (WriteFile(fileHandle, memory, memorySize, &bytesWritten, NULL)) {
            // file written successfully
            result = bytesWritten == memorySize;
        } else {
            // logging
        }
        CloseHandle(fileHandle);
    } else {
        // logging
    }
    return result;
}

internal void resizeDIBSection(Win32OffscreenBuffer *buffer, int width, int height) {
    if (buffer->memory) {
        VirtualFree(buffer->memory, NULL, MEM_RELEASE);
    }

    buffer->width = width;
    buffer->height = height;
    buffer->bytesPerPixel = 4;

    buffer->info.bmiHeader.biSize = sizeof(buffer->info.bmiHeader);
    buffer->info.bmiHeader.biWidth = buffer->width;
    buffer->info.bmiHeader.biHeight = -buffer->height;
    buffer->info.bmiHeader.biPlanes = 1;
    buffer->info.bmiHeader.biBitCount = 32;
    buffer->info.bmiHeader.biCompression = BI_RGB;

    int bitmapSize = buffer->width*buffer->height*4;
    buffer->memory = VirtualAlloc(NULL, bitmapSize, MEM_COMMIT, PAGE_READWRITE);
}

internal Win32WindowDimensions win32GetWindowDimensions(HWND window) {
    Win32WindowDimensions result;
    RECT clientRect;
    GetClientRect(window, &clientRect);
    result.width = clientRect.right - clientRect.left;
    result.height = clientRect.bottom - clientRect.top;
    return result;
}

internal void win32DrawBufferInWindow(Win32OffscreenBuffer *buffer, HWND window) {
    Win32WindowDimensions winDims = win32GetWindowDimensions(window);
    StretchDIBits(
        GetDC(window),
        0, 0, winDims.width, winDims.height,
        0, 0, buffer->width, buffer->height,
        buffer->memory,
        &buffer->info, 
        DIB_RGB_COLORS,
        SRCCOPY
    );
}

internal void win32LoadXInput() {
    HMODULE xInputLib = LoadLibrary("xinput1_4.dll");
    if (!xInputLib) {
        xInputLib = LoadLibrary("xinput1_3.dll");
    }

    if (xInputLib) {
        XInputGetState = (XInputGetStateFn *)GetProcAddress(xInputLib, "XInputGetState");
        XInputSetState = (XInputSetStateFn *)GetProcAddress(xInputLib, "XInputSetState");
    }
}

internal void win32ProcessKeyboardKeypress(GameButtonState *newState, bool32 isDown) {
    newState->endedDown = isDown;
    newState->halfTransitionCount++;
}

internal void win32ProcessXInputDigitalButton(DWORD xInputButtonState, GameButtonState *oldState, GameButtonState *newState, DWORD buttonBit) {
    newState->endedDown = (xInputButtonState & buttonBit) == buttonBit;
    newState->halfTransitionCount = (oldState->endedDown != newState->endedDown) ? 1 : 0;
}

internal void win32InitSound(HWND window, int32 samplesPerSec, int bufferSize) {
    HMODULE dSoundLibrary = LoadLibraryA("dsound.dll");
    if (dSoundLibrary) {
        DirectSoundCreateFn *DirectSoundCreate = (DirectSoundCreateFn*)GetProcAddress(dSoundLibrary, "DirectSoundCreate");
        LPDIRECTSOUND DirectSound;
        if (DirectSoundCreate && SUCCEEDED(DirectSoundCreate(0, &DirectSound, 0))) {
            WAVEFORMATEX waveFormat = {};
            waveFormat.wFormatTag = WAVE_FORMAT_PCM;
            waveFormat.nChannels = 2;
            waveFormat.nSamplesPerSec = samplesPerSec;
            waveFormat.wBitsPerSample = 16;
            waveFormat.nBlockAlign = (waveFormat.nChannels*waveFormat.wBitsPerSample) / 8;
            waveFormat.nAvgBytesPerSec = waveFormat.nSamplesPerSec*waveFormat.nBlockAlign;
            waveFormat.cbSize = 0;

            if (SUCCEEDED(DirectSound->SetCooperativeLevel(window, DSSCL_PRIORITY))) {
                DSBUFFERDESC bufferDescription = {};
                bufferDescription.dwSize = sizeof(bufferDescription);
                bufferDescription.dwFlags = DSBCAPS_PRIMARYBUFFER;

                LPDIRECTSOUNDBUFFER primaryBuffer;
                if (SUCCEEDED(DirectSound->CreateSoundBuffer(&bufferDescription, &primaryBuffer, 0))) {
                    if (SUCCEEDED(primaryBuffer->SetFormat(&waveFormat))) {
                        OutputDebugStringA("Primary buffer format was set.\n");
                    } else {
                        //No sound! (?)
                    }
                }
            } else {
                // No sound! (?)
            }

            DSBUFFERDESC bufferDescription = {};
            bufferDescription.dwSize = sizeof(bufferDescription);
            bufferDescription.dwFlags = 0;
            bufferDescription.dwBufferBytes = bufferSize;
            bufferDescription.lpwfxFormat = &waveFormat;

            if (SUCCEEDED(DirectSound->CreateSoundBuffer(&bufferDescription, &globalSecondaryBuffer, 0))) {
                OutputDebugStringA("Secondary buffer created successfully!\n");
            }
        } else {
            // No sound! (?)
        }
    }
}

LRESULT mainWindowCallback(
  HWND window,
  UINT message,
  WPARAM wParam,
  LPARAM lParam
) {
    LRESULT result = 0;
    switch (message) {
        case WM_SIZE: {
        } break;

        case WM_DESTROY: {
            OutputDebugStringA("WM_DESTROY\n");
        } break;

        case WM_CLOSE: {
            globalRunning = false;
        } break;

        case WM_PAINT: {
            PAINTSTRUCT paint = {};
            HDC deviceContext = BeginPaint(window, &paint);
            int x = paint.rcPaint.left;
            int y = paint.rcPaint.top;
            int width = paint.rcPaint.right - paint.rcPaint.left;
            int height = paint.rcPaint.bottom - paint.rcPaint.top;
            win32DrawBufferInWindow(&globalBackBuffer, window);
            EndPaint(window, &paint);
        } break;

        case WM_ACTIVATEAPP: {
        } break;

        case WM_HOTKEY: {
            if (wParam == HH_CTRLW) {
                globalRunning = false;
            }
        } break;

        case WM_SYSKEYDOWN:
        case WM_SYSKEYUP:
        case WM_KEYUP:
        case WM_KEYDOWN: {
            Assert(!"Keyboard input came in through a non-dispatch message.");
        } break;

        default: {
            result = DefWindowProcA(window, message, wParam, lParam);
        } break;
    }
    return result;
}

internal void win32ClearBuffer(Win32SoundOutput *soundOutput) {
    VOID *region1;
    DWORD region1Size;
    VOID *region2;
    DWORD region2Size;
    HRESULT lock = globalSecondaryBuffer->Lock(0, soundOutput->secondaryBufferSize, &region1, &region1Size, &region2, &region2Size, 0);
    if (SUCCEEDED(lock)) {
        int16 *destSamples = (int16*)region1;
        for (DWORD byteIndex = 0; byteIndex < region1Size; byteIndex++) {
           // *destSamples++ = 0;
        }
        destSamples = (int16*)region2;
        for (DWORD byteIndex = 0; byteIndex < region2Size; byteIndex++) {
            //*destSamples++ = 0;
        }
        globalSecondaryBuffer->Unlock(region1, region1Size, region2, region2Size);
    } else {
        // no lock!
    }
}

internal void win32FillSoundBuffer(Win32SoundOutput *soundOutput, DWORD byteToLock, DWORD bytesToWrite, GameSoundOutputBuffer *sourceBuffer) {
    VOID *region1;
    DWORD region1Size;
    VOID *region2;
    DWORD region2Size;
    HRESULT lock = globalSecondaryBuffer->Lock(byteToLock, bytesToWrite, &region1, &region1Size, &region2, &region2Size, 0);
    if (SUCCEEDED(lock)) {
        DWORD region1SampleCount = region1Size/soundOutput->bytesPerSample;
        int16 *destSamples = (int16*)region1;
        int16 *sourceSamples = sourceBuffer->samples;
        for (DWORD sampleIndex = 0; sampleIndex < region1SampleCount; sampleIndex++) {
            *destSamples++ = *sourceSamples++;
            *destSamples++ = *sourceSamples++;
            soundOutput->runningSampleIndex++;
        }
        DWORD region2SampleCount = region2Size/soundOutput->bytesPerSample;
        destSamples = (int16*)region2;
        for (DWORD sampleIndex = 0; sampleIndex < region2SampleCount; sampleIndex++) {
            *destSamples++ = *sourceSamples++;
            *destSamples++ = *sourceSamples++;
            soundOutput->runningSampleIndex++;
        }
        globalSecondaryBuffer->Unlock(region1, region1Size, region2, region2Size);
    } else {
        // no lock!
    }
}

internal void win32ProcessPendingMessages(GameControllerInput *keyboardController) {
    MSG message;
    while (PeekMessageA(&message, NULL, NULL, NULL, PM_REMOVE)) {
        if (message.message == WM_QUIT) {
            globalRunning = false;
        }

        switch (message.message) {
            case WM_SYSKEYDOWN:
            case WM_SYSKEYUP:
            case WM_KEYUP:
            case WM_KEYDOWN: {
                uint32 VKCode = (uint32)message.wParam;
                bool wasDown = (message.lParam & (1 << 30)) != 0;
                bool isDown = (message.lParam & (1 << 31)) == 0;
                if (wasDown != isDown) {
                    if (VKCode == 'W') {
                        win32ProcessKeyboardKeypress(&keyboardController->stickUp, isDown);
                    } else if (VKCode == 'A') {
                        win32ProcessKeyboardKeypress(&keyboardController->stickLeft, isDown);
                    } else if (VKCode == 'S') {
                        win32ProcessKeyboardKeypress(&keyboardController->stickDown, isDown);
                    } else if (VKCode == 'D') {
                        win32ProcessKeyboardKeypress(&keyboardController->stickRight, isDown);
                    } else if (VKCode == 'Q') {
                        win32ProcessKeyboardKeypress(&keyboardController->leftShoulder, isDown);
                    } else if (VKCode == 'E') {
                        win32ProcessKeyboardKeypress(&keyboardController->rightShoulder, isDown);
                    } else if (VKCode == VK_ESCAPE) {
                    } else if (VKCode == VK_UP) {
                        win32ProcessKeyboardKeypress(&keyboardController->btnUp, isDown);
                    } else if (VKCode == VK_LEFT) {
                        win32ProcessKeyboardKeypress(&keyboardController->btnLeft, isDown);
                    } else if (VKCode == VK_DOWN) {
                        win32ProcessKeyboardKeypress(&keyboardController->btnDown, isDown);
                    } else if (VKCode == VK_RIGHT) {
                        win32ProcessKeyboardKeypress(&keyboardController->btnRight, isDown);
                    } else if (VKCode == VK_SPACE) {
                    }
                }

                bool32 altKeyWasDown = message.lParam & (1 << 29);
                if (altKeyWasDown && VKCode == VK_F4) {
                    globalRunning = false;
                }
            } break;

            default: {
                TranslateMessage(&message);
                DispatchMessage(&message);
            } break;
        }
    }
}

inline LARGE_INTEGER win32GetWallClock() {
    LARGE_INTEGER result;
    QueryPerformanceCounter(&result);
    return result;
}

inline real32 win32GetSecondsElapsed(LARGE_INTEGER start, LARGE_INTEGER end) {
    return (real32)(end.QuadPart - start.QuadPart) / (real32)globalPerfCountFrequency;
}

//    int monitorRefreshHz = 60;
//    int gameUpdateHz = monitorRefreshHz / 2;

#define monitorRefreshHz 60
#define gameUpdateHz (monitorRefreshHz / 2)

internal void win32DebugDrawVertical(Win32OffscreenBuffer *screenBuffer, int x, int top, int bottom, int32 color) {
    int pitch = screenBuffer->width*screenBuffer->bytesPerPixel; 
    uint8 *pixel = (uint8 *)screenBuffer->memory + top * pitch + x*screenBuffer->bytesPerPixel;
    for (int y = top; y < bottom; y++) {
        *(uint32 *)pixel = color;
        pixel += pitch;
    }
}

inline void win32DrawSoundBufferMarker(Win32OffscreenBuffer *buffer, Win32SoundOutput *soundOutput, real32 pxPerSoundBufferEntry, int padX, int top, int bottom, DWORD value, uint32 color) {
    Assert(value < (DWORD)soundOutput->secondaryBufferSize);
    int x = padX + (int)(pxPerSoundBufferEntry * (real32)value);
    win32DebugDrawVertical(buffer, x, top, bottom, color);
}

internal void win32DebugSyncDisplay(Win32OffscreenBuffer *screenBuffer, int markerCount, Win32DebugTimeMarker *markers, Win32SoundOutput *soundOutput, real32 targetSecondsPerFrame) {
    int padX = 16;
    int padY = 16;

    int top = padY;
    int bottom = screenBuffer->height - padY;
    int renderWidth = screenBuffer->width - 2 *padX;

    real32 pxPerSoundBufferEntry = (real32)renderWidth / (real32)soundOutput->secondaryBufferSize;
    int pitch = screenBuffer->width*screenBuffer->bytesPerPixel; 

    for (int markerIndex = 0; markerIndex < markerCount; markerIndex++) {
        Win32DebugTimeMarker *thisMarker = &markers[markerIndex];
        real32 alpha = ((real32)(markerIndex + 1) / (real32)markerCount);
        int x = padX + (int)(pxPerSoundBufferEntry * (real32)thisMarker->writeCursor);
        uint8 *pixel = (uint8 *)screenBuffer->memory + top * pitch + x*screenBuffer->bytesPerPixel;
        uint32 newPixel = (uint32)((real32)alpha * 0xFFFFFFFF + (1.0f - alpha) * (*(uint32 *)pixel));
        win32DrawSoundBufferMarker(screenBuffer, soundOutput, pxPerSoundBufferEntry, padX, top, bottom, thisMarker->playCursor, newPixel);
        win32DrawSoundBufferMarker(screenBuffer, soundOutput, pxPerSoundBufferEntry, padX, top, bottom, thisMarker->writeCursor, 0x00FF0000);
    }
}

int APIENTRY WinMain(HINSTANCE instance, HINSTANCE prevInstance, PSTR commandLine, int commandShow) {
    LARGE_INTEGER performanceFrequencyResult;
    QueryPerformanceFrequency(&performanceFrequencyResult);
    globalPerfCountFrequency = performanceFrequencyResult.QuadPart;

    UINT desiredSchedulerMs = 1;
    bool32 sleepIsGranular = TIMERR_NOERROR == timeBeginPeriod(desiredSchedulerMs);

    WNDCLASSA windowClass = {};
    windowClass.style = CS_VREDRAW | CS_HREDRAW;
    windowClass.lpfnWndProc = &mainWindowCallback;
    windowClass.hInstance = instance;
    windowClass.lpszClassName = "HandmadeHeroWindowClass";

    resizeDIBSection(&globalBackBuffer, 1280, 720);

    real32 targetSecondsPerFrame = 1.0f / (real32)gameUpdateHz;

    if (RegisterClass(&windowClass)) {
        HWND window = CreateWindowExA(
            NULL,
            windowClass.lpszClassName,
            "Handmade Hero",
            WS_OVERLAPPEDWINDOW | WS_VISIBLE,
            CW_USEDEFAULT,
            CW_USEDEFAULT,
            CW_USEDEFAULT,
            CW_USEDEFAULT,
            NULL,
            NULL,
            instance,
            NULL
        );
        if (window) {
            HH_CTRLW = GlobalAddAtomA("HH_CTRLW");
            RegisterHotKey(window, HH_CTRLW, MOD_CONTROL, 'W');

            globalRunning = true;
            GameInput input[2] = {};
            GameInput *oldInput = &input[0];
            GameInput *newInput = &input[1];

            int debugTimeMarkerIndex = 0; 
            Win32DebugTimeMarker debugMarkers[gameUpdateHz / 2] = {};

            win32LoadXInput();

            // sound test
            Win32SoundOutput soundOutput = {};
            soundOutput.samplesPerSecond = 48000;
            soundOutput.runningSampleIndex = 0;
            soundOutput.bytesPerSample = sizeof(int16)*2;
            soundOutput.secondaryBufferSize = soundOutput.samplesPerSecond*soundOutput.bytesPerSample;
            soundOutput.safetyBytes = (soundOutput.samplesPerSecond * soundOutput.bytesPerSample / gameUpdateHz) / 3;

            int16 *samples = (int16*)VirtualAlloc(NULL, soundOutput.secondaryBufferSize, MEM_COMMIT, PAGE_READWRITE);

#if HANDMADE_INTERNAL
            LPVOID baseAddress = (LPVOID)Terabytes((uint64)2);
#else
            LPVOID baseAddress = 0;
#endif
            GameMemory gameMemory = {};
            gameMemory.permanentStorageSize = Megabytes(64);
            gameMemory.transientStorageSize = Gigabytes((uint64)4);

            uint64 totalSize = gameMemory.permanentStorageSize + gameMemory.transientStorageSize;
            gameMemory.permanentStorage = VirtualAlloc(baseAddress, totalSize, MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE);
            gameMemory.transientStorage = ((uint8 *)gameMemory.permanentStorage + gameMemory.permanentStorageSize);

            win32InitSound(window, soundOutput.samplesPerSecond, soundOutput.secondaryBufferSize);
            win32ClearBuffer(&soundOutput);
            globalSecondaryBuffer->Play(0, 0, DSBPLAY_LOOPING);

            LARGE_INTEGER lastWorkCounter = win32GetWallClock();
            LARGE_INTEGER flipWallClock = win32GetWallClock();

            int64 lastCycleCount = __rdtsc();

            DWORD audioLatencyBytes = 0;
            real32 audioLatencySeconds = 0;

            while (globalRunning) {
                GameControllerInput *oldKeyboardController = &oldInput->controllers[0];
                GameControllerInput *newKeyboardController = &newInput->controllers[0];
                GameControllerInput zeroController = {};
                *newKeyboardController = zeroController;
                for (int buttonIndex = 0; buttonIndex < ArrayCount(oldKeyboardController->buttons); buttonIndex++) {
                    newKeyboardController->buttons[buttonIndex].endedDown = oldKeyboardController->buttons[buttonIndex].endedDown;
                }

                win32ProcessPendingMessages(newKeyboardController);

                XINPUT_STATE controllerState;
                int maxControllerCount = XUSER_MAX_COUNT;
                if (maxControllerCount > ArrayCount(newInput->controllers) - 1) {
                    maxControllerCount = ArrayCount(newInput->controllers) - 1;
                }

                for (DWORD controllerIndex = 0; controllerIndex < XUSER_MAX_COUNT; controllerIndex++) {
                    if (XInputGetState(controllerIndex, &controllerState) == ERROR_SUCCESS) {
                        DWORD ourControllerIndex = controllerIndex + 1;
                        GameControllerInput *oldController = &oldInput->controllers[ourControllerIndex];
                        GameControllerInput *newController = &newInput->controllers[ourControllerIndex];

                        newController->isAnalog = true;

                        XINPUT_GAMEPAD *pad = &controllerState.Gamepad;

                        int16 stickX = pad->sThumbLX;
                        int16 stickY = pad->sThumbLY;
                        real32 stickYNorm = 0;
                        real32 stickXNorm = 0;
                        if (stickY != 0 && abs(stickY) > XINPUT_GAMEPAD_LEFT_THUMB_DEADZONE) {
                            stickYNorm = stickY / 32767.0f;
                        }
                        if (stickX != 0 && abs(stickX) > XINPUT_GAMEPAD_LEFT_THUMB_DEADZONE) {
                            stickXNorm = stickX / 32767.0f;
                        }

                        newController->stickAvgX = stickXNorm;
                        newController->stickAvgY = stickYNorm;
                        if (pad->wButtons & XINPUT_GAMEPAD_DPAD_UP) {
                            newController->stickAvgY = 1.0f;
                        }
                        if (pad->wButtons & XINPUT_GAMEPAD_DPAD_DOWN) {
                            newController->stickAvgY = -1.0f;
                        }
                        if (pad->wButtons & XINPUT_GAMEPAD_DPAD_LEFT) {
                            newController->stickAvgX = -1.0f;
                        }
                        if (pad->wButtons & XINPUT_GAMEPAD_DPAD_RIGHT) {
                            newController->stickAvgX = 1.0f;
                        }

                        //bool start = (pad->wButtons & XINPUT_GAMEPAD_START);
                        //bool back = (pad->wButtons & XINPUT_GAMEPAD_BACK);

                        win32ProcessXInputDigitalButton(pad->wButtons, &oldController->btnDown, &newController->btnDown, XINPUT_GAMEPAD_A);
                        win32ProcessXInputDigitalButton(pad->wButtons, &oldController->btnRight, &newController->btnRight, XINPUT_GAMEPAD_B);
                        win32ProcessXInputDigitalButton(pad->wButtons, &oldController->btnLeft, &newController->btnLeft, XINPUT_GAMEPAD_X);
                        win32ProcessXInputDigitalButton(pad->wButtons, &oldController->btnUp, &newController->btnUp, XINPUT_GAMEPAD_Y);
                        win32ProcessXInputDigitalButton(pad->wButtons, &oldController->leftShoulder, &newController->leftShoulder, XINPUT_GAMEPAD_LEFT_SHOULDER);
                        win32ProcessXInputDigitalButton(pad->wButtons, &oldController->rightShoulder, &newController->rightShoulder, XINPUT_GAMEPAD_RIGHT_SHOULDER);

                        real32 threshold = 0.5f;
                        win32ProcessXInputDigitalButton((DWORD)(newController->stickAvgY < -threshold), &oldController->stickDown, &newController->stickDown, 1);
                        win32ProcessXInputDigitalButton((DWORD)(newController->stickAvgX > threshold), &oldController->stickRight, &newController->stickRight, 1);
                        win32ProcessXInputDigitalButton((DWORD)(newController->stickAvgX < -threshold), &oldController->stickLeft, &newController->stickLeft, 1);
                        win32ProcessXInputDigitalButton((DWORD)(newController->stickAvgY > threshold), &oldController->stickUp, &newController->stickUp, 1);
                    } else {
                        // controller not available
                    }
                }

                GameOffscreenBuffer videoBuffer = {};
                videoBuffer.memory = globalBackBuffer.memory;
                videoBuffer.width = globalBackBuffer.width;
                videoBuffer.height = globalBackBuffer.height;
                gameUpdateAndRender(&gameMemory, &videoBuffer, newInput);

                DWORD playCursor = 0;
                DWORD writeCursor = 0;
                bool soundIsValid = false;

                if (SUCCEEDED(globalSecondaryBuffer->GetCurrentPosition(&playCursor, &writeCursor))) {
                    if (!soundIsValid) {
                        soundOutput.runningSampleIndex = writeCursor / soundOutput.bytesPerSample;
                        soundIsValid = true;
                    }
                    DWORD byteToLock = (soundOutput.runningSampleIndex*soundOutput.bytesPerSample) % soundOutput.secondaryBufferSize;

                    // TODO(dledda): episode 20 ~2:00, something's wrong though, gotta sort it out.  

                    DWORD expectedSoundBytesPerFrame = (soundOutput.samplesPerSecond * soundOutput.bytesPerSample) / gameUpdateHz;
                    DWORD expectedFrameBoundaryByte = playCursor + expectedSoundBytesPerFrame;
                    DWORD safeWriteCursor = writeCursor;
                    if (safeWriteCursor < playCursor) {
                        safeWriteCursor += soundOutput.secondaryBufferSize;
                    }
                    Assert(safeWriteCursor >= playCursor);
                    safeWriteCursor += soundOutput.safetyBytes + 2000;
                    bool audioIsLowLatency = safeWriteCursor >= expectedFrameBoundaryByte;

                    DWORD targetCursor = 0;
                    if (audioIsLowLatency) {
                        targetCursor = expectedFrameBoundaryByte + expectedSoundBytesPerFrame;
                    } else {
                        targetCursor = writeCursor + expectedSoundBytesPerFrame + soundOutput.safetyBytes;
                    }
                    targetCursor %= soundOutput.secondaryBufferSize;

                    DWORD bytesToWrite = 0;
                    if (byteToLock > targetCursor) {
                        bytesToWrite = soundOutput.secondaryBufferSize - byteToLock + targetCursor;
                    } else {
                        bytesToWrite = targetCursor - byteToLock;
                    }

                    GameSoundOutputBuffer soundBuffer = {};
                    soundBuffer.samplesPerSecond = soundOutput.samplesPerSecond;
                    soundBuffer.sampleCount = bytesToWrite / soundOutput.bytesPerSample;
                    soundBuffer.samples = samples;
                    gameGetSoundSamples(&gameMemory, &soundBuffer);
#if HANDMADE_INTERNAL
                    Assert(debugTimeMarkerIndex < ArrayCount(debugMarkers));

                    Win32DebugTimeMarker *marker = &debugMarkers[debugTimeMarkerIndex++];

                    DWORD unwrappedWriteCursor = writeCursor;
                    if (unwrappedWriteCursor < targetCursor) {
                        unwrappedWriteCursor += soundOutput.secondaryBufferSize;
                    }
                    audioLatencyBytes = unwrappedWriteCursor - playCursor;
                    audioLatencySeconds = (((real32)audioLatencyBytes / (real32)soundOutput.bytesPerSample) / (real32)soundOutput.samplesPerSecond);

                    if (debugTimeMarkerIndex >= ArrayCount(debugMarkers)) {
                        debugTimeMarkerIndex = 0;
                    }
                    
                    globalSecondaryBuffer->GetCurrentPosition(&marker->playCursor, &marker->writeCursor);
#endif
                    win32FillSoundBuffer(&soundOutput, byteToLock, bytesToWrite, &soundBuffer);
                } else {
                    soundIsValid = false;
                }


                real32 secondsElapsedForFrame = win32GetSecondsElapsed(lastWorkCounter, win32GetWallClock());
                if (secondsElapsedForFrame < targetSecondsPerFrame) {
                    if (sleepIsGranular) {
                        DWORD sleepMs = (DWORD)(1000.0f * (targetSecondsPerFrame - secondsElapsedForFrame));
                        if (sleepMs > 0) {
                            Sleep(sleepMs);
                        }
                    }
                    while (secondsElapsedForFrame < targetSecondsPerFrame) {
                        secondsElapsedForFrame = win32GetSecondsElapsed(lastWorkCounter, win32GetWallClock());
                    }
                } else {
                    // logging, MISSED FRAME!
                }

                lastWorkCounter = win32GetWallClock();

#if HANDMADE_INTERNAL
                win32DebugSyncDisplay(&globalBackBuffer, ArrayCount(debugMarkers), debugMarkers, &soundOutput, targetSecondsPerFrame);
#endif

                win32DrawBufferInWindow(&globalBackBuffer, window);
                flipWallClock = win32GetWallClock();

#if 0
                real64 msElapsed = (real32)(1000.0f*secondsElapsedForFrame) / globalPerfCountFrequency;
                real64 fps = (real32)(1000.0f*globalPerfCountFrequency/(real32)secondsElapsedForFrame)/1000.0f;
                int64 endCycleCount = __rdtsc();
                int64 cyclesElapsed = endCycleCount - lastCycleCount;
                lastCycleCount = endCycleCount;
#endif

                GameInput *temp = newInput;
                newInput = oldInput; 
                oldInput = temp;
            }
        } else {
            // failed
        }
    } else {
        // failed
    }

    return 0;
}

